# Utils for creating swaptions

import numpy as np
from qablet_contracts.timetable import timetable_from_dicts


def simple_swap_period(
    ccy: str, start: float, end: float, fixed_rate: float, track: str = ""
) -> list:
    """Simple representation of a swap period, paying fixed, receiving floating rate.

    Args:
        ccy: the currency of the swap.
        start: the start of the period in years.
        end: the end of the period in years.
        fixed_rate: the fixed annual rate of the swap.
        track: an optional identifier for the contract.
    """
    return [
        {
            "track": track,
            "time": start,
            "op": "+",
            "quantity": 1,
            "unit": ccy,
        },
        {
            "track": track,
            "time": end,
            "op": "+",
            "quantity": -1 - fixed_rate * (end - start),
            "unit": ccy,
        },
    ]


def bermuda_swaption_timetable(
    ccy: str, times: list[float], strike_rate: float, track: str = ""
) -> dict:
    """Create timetable for a co-terminal bermuda swaption.

    Args:
        ccy: the currency of the swap.
        times: the period times of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.
    """

    events = []
    for start, end in zip(times[0:-1], times[1:]):
        # option expiration event before each period
        events.append(
            {
                "track": track + ".opt",
                "time": start,
                "op": ">",
                "quantity": 1,
                "unit": track + ".swp",
            }
        )
        # payment event for the underlying swap
        events.extend(
            simple_swap_period(ccy, start, end, strike_rate, track + ".swp")
        )

    return timetable_from_dicts(events)


def swaption_timetable(
    ccy: str, times: list[float], strike_rate: float, track: str = ""
) -> dict:
    """Create timetable for a vanilla swaption.

    Args:
        ccy: the currency of the swap.
        times: the period times of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.
    """

    # option expiration event at beginning of the swap
    events = [
        {
            "track": track + ".opt",
            "time": times[0],
            "op": ">",
            "quantity": 1,
            "unit": track + ".swp",
        }
    ]
    # payment events for the underlying swap
    for start, end in zip(times[0:-1], times[1:]):
        events.extend(
            simple_swap_period(ccy, start, end, strike_rate, track + ".swp")
        )

    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create the bermuda swaption
    times = np.linspace(
        1, 6, 6
    ).tolist()  # Start at 1 year, mature at 6 years.
    strike_rate = 0.03
    timetable = bermuda_swaption_timetable("USD", times, strike_rate)

    print(timetable["events"].to_pandas())



"""
Utils for creating cliquet contracts.
"""

import numpy as np
import pyarrow as pa
from qablet_contracts.timetable import EVENT_SCHEMA


def clique_timetable(
    ccy: str,
    asset_name: str,
    fixings: list[float],
    global_floor: float,
    local_floor: float,
    local_cap: float,
    track: str = "",
) -> dict:
    """Create timetable for an accumulator cliquet.

    Args:
        ccy: the currency of the bond.
        asset_name: the name of the underlying asset.
        fixings: the fixing times of the cliquet.
        global_floor: the global floor of the cliquet.
        local_floor: the local floor of the cliquet.
        local_cap: the local cap of the cliquet.
        track: an optional identifier for the contract.
    """

    maturity = fixings[-1]

    # define timetable
    events = [
        {
            "track": None,
            "time": fixings[0],
            "op": "s",
            "quantity": 0,
            "unit": "_INIT",  # initialize accumulator
        }
    ]
    for fixing_time in fixings[1:]:
        events.append(
            {
                "track": None,
                "time": fixing_time,
                "op": "s",
                "quantity": 0,
                "unit": "_UPDATE",  # update accumulator
            }
        )
    events.append(
        {
            "track": track,
            "time": maturity,
            "op": ">",  # global floor
            "quantity": global_floor,
            "unit": ccy,
        }
    )
    events.append(
        {
            "track": track,
            "time": maturity,
            "op": "+",  # pay the accumulated amount
            "quantity": 1,
            "unit": "_A",
        }
    )

    # define accumulator functions
    def accumulator_init_fn(inputs):
        [s] = inputs
        return [0, s]  # [A, S_last]

    def accumulator_update_fn(inputs):
        [s, s_last, a] = inputs

        ret = s / s_last - 1.0  # ret = S / S_last - 1
        ret = np.maximum(local_floor, ret)
        ret = np.minimum(local_cap, ret)

        return [a + ret, s]  # [A, S_last]

    return {
        "events": pa.RecordBatch.from_pylist(events, schema=EVENT_SCHEMA),
        "expressions": {
            "_INIT": {
                "type": "snapper",
                "inp": [asset_name],
                "snap_fn": accumulator_init_fn,
                "out": ["_A", "_S_last"],
            },
            "_UPDATE": {
                "type": "snapper",
                "inp": [asset_name, "_S_last", "_A"],
                "snap_fn": accumulator_update_fn,
                "out": ["_A", "_S_last"],
            },
        },
    }


if __name__ == "__main__":
    # Create the cliquet
    global_floor = 0.0
    fixings = np.linspace(0, 3, 7).tolist()  # T = 3 years, N = 6 fixings
    local_floor = -0.03
    local_cap = 0.05
    timetable = clique_timetable(
        "USD", "SPX", fixings, global_floor, local_floor, local_cap
    )

    print(timetable["events"].to_pandas())



"""
Utils for creating vanilla options timetable
"""

from qablet_contracts.timetable import timetable_from_dicts


def option_timetable(
    ccy: str,
    asset_name: str,
    strike: float,
    maturity: float,
    is_call: bool,
    track: str = "",
) -> dict:
    """Create timetable for an equity vanilla call option.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike: the option strike.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.
    """

    events = [
        {
            "track": track,
            "time": maturity,
            "op": ">",
            "quantity": 0,
            "unit": ccy,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": -strike if is_call else strike,
            "unit": ccy,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": 1 if is_call else -1,
            "unit": asset_name,
        },
    ]
    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create the option
    timetable = option_timetable("USD", "SPX", 2900, 0.5, True)
    print(timetable["events"].to_pandas())



"""
This module contains examples of creating timetables for Zero Coupon Bond and related contracts.
"""

from qablet_contracts.timetable import timetable_from_dicts


def zcb_timetable(ccy: str, maturity: float, track: str = "") -> dict:
    """Create timetable for a zero coupon bond.

    Args:
        ccy: the currency of the bond.
        maturity: the maturity of the bond in years.
        track: an optional identifier for the contract.
    """
    events = [
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": 1,
            "unit": ccy,
        }  # get bond notional at bond expiration
    ]
    return timetable_from_dicts(events)


def zbp_timetable(
    ccy: str,
    opt_maturity: float,
    bond_maturity: float,
    strike: float,
    track: str = "",
) -> dict:
    """Create timetable for a zero coupon bond put.

    Args:
        ccy: the currency of the bond.
        opt_maturity: the maturity of the option in years.
        bond_maturity: the maturity of the option in years.
        strike: the option strike.
        track: an optional identifier for the contract.
    """

    events = [
        {
            "track": track,
            "time": opt_maturity,
            "op": ">",
            "quantity": 0,
            "unit": ccy,
        },  # Choose greater of nothing (get 0) or exercise (continue to remaining events)
        {
            "track": track,
            "time": opt_maturity,
            "op": "+",
            "quantity": strike,
            "unit": ccy,
        },  # get strike at expiration
        {
            "track": track,
            "time": bond_maturity,
            "op": "+",
            "quantity": -1,
            "unit": ccy,
        },  # pay bond notional at bond expiration
    ]
    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create a zero coupon bond timetable
    timetable = zcb_timetable("USD", 1)
    print("zcb:\n", timetable["events"].to_pandas())

    timetable = zbp_timetable("USD", 0.5, 1.0, 0.95)
    print("zbp:\n", timetable["events"].to_pandas())



# Define the timetable struct

import pyarrow as pa

DICT_TYPE = pa.dictionary(pa.int64(), pa.string())
EVENT_SCHEMA = pa.schema(
    [
        pa.field("track", DICT_TYPE),
        pa.field("time", pa.float64()),
        pa.field("op", DICT_TYPE),
        pa.field("quantity", pa.float64()),
        pa.field("unit", DICT_TYPE),
    ]
)


def timetable_from_dicts(events: list[dict]) -> dict:
    """Create timetable from a list of dicts.

    Args:
        events: a list of dicts with the following fields:

            - track: string
            - time: float
            - op: string
            - quantity: float
            - unit: string

    Returns:
        a timetable dictionary with the following fields:

            - events: a pyarrow record batch
            - expressions: a dict for expressions, batches, and snappers

    """
    return {
        "events": pa.RecordBatch.from_pylist(events, schema=EVENT_SCHEMA),
        "expressions": {},
    }
